%% tclp__files.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__files.pl,v 1.9 2005/05/26 08:51:04 coquery Exp $

:- module(tclp__files,[ % read_file/2, load_file/1,
                        load_now/1,
                        load_from_lines/2,
                        read_from_lines/3,
                        add_include/1,
                        chop_ext/3,
                        print_file/1,
                        tclp__directory/1,
                        set_tclp__directory/1,
                        expand_path/2,
                        add_file_to_read/1,
                        add_file_to_load/1,
                        read_and_load/1,
                        file_in_tclp_path/1,
                        tclp__consult/1
                      ]).

:- use_module(library(lists)).
:- use_module(library(charsio)).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__handle_goals)).
:- use_module(library(system)).
:- use_module(library(chr)).
:- use_module(library(system)).

%% Verbose mode : debug_files

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% File names
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% We add the possibility to have including directories
%% Include dirs are stored in a dynamic predicate

:- dynamic include_dir/1.

%include_dir('../').

%% The predicate to add a directory to include dirs
add_include(X) :-
        absolute_file_name(X,Y),
        assertz(include_dir(Y)).

%% last_in_list(X,List) is true if X can be unified with the last element of
%% list
last_in_list(X,[X]) :- !.
last_in_list(X,[_|Rest]) :- last_in_list(X,Rest).

%% concat_filename(DirName, FileName, CompleteFileName)
%% CompleteFileName is DirName/FileName with any trailing / removed from
%% DirName
concat_filename(DirName, FileName, CompleteFileName) :-
        atom_concat(_,'/',DirName),!, 
        atom_concat(DirName, FileName, CompleteFileName).
concat_filename(DirName, FileName, CompleteFileName) :-
        atom_concat(DirName,'/', CDirName),
        atom_concat(CDirName,FileName, CompleteFileName).

%% Enumerate possible file names
filename(FileName, CompleteFileName) :-
        FileName = CompleteFileName ;
        include_dir(Dir),
        concat_filename(Dir, FileName, CompleteFileName).

%% chop_ext(FileName, Ext, Chopped)
%% true if FileName ends with Ext and ChoppedExt = FileName
chop_ext(FileName, Ext, Chopped) :-
        atom_concat(Chopped, Ext, FileName).

:- multifile tclp__files:file_search_path/2.
:- dynamic tclp__files:file_search_path/2.
%% expand the file name using file_search_path/2
expand_path(Path,Path) :- atomic(Path),!.
expand_path(Path,Expanded) :-
        Path =.. [ Prefix , Rest ],
        expand_path(Rest,Rest2),
        tclp__files:file_search_path(Prefix,ExpandedPrefix),
        concat_filename(ExpandedPrefix,Rest2,Expanded).

%% true if the given file can be found in the current tclp path for files
file_in_tclp_path(FileName) :-
        filename(FileName, CompleteFileName),
        file_exists(CompleteFileName).
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% File reading / loading
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% open the file by looking into include directories
open_stream(FileName, Stream) :-
        (
          FileName == '-'
        ->
          Stream = user_input
        ;
          (
            nofileerrors,
            filename(FileName, CompleteFileName),
            verbose(debug_files,['Trying to open ', CompleteFileName]),
            open(CompleteFileName, read, Stream),!
          ;
            %% print_error([ 'Can''t open "', FileName, '"' ])
            throw(filenotfound(FileName))
          )
        ),
        verbose(debug_files,'stream openned').

print_file(FileName) :-
        catch(open_stream(FileName, Stream),
              filenotfound(F),
              print_error([ 'Can''t open "', F, '"' ])),
        repeat,
        get_code(Stream,Code),
        (
          Code = -1
        ->
          true
        ;
          put_code(Code),
          fail
        ),!,
        close(Stream).

%% Read a stream : read each phrase of the stream, accumulating lines number
%% and variable names.
%% also executes directives when needed
read_stream(FileName, Stream, PhraseList) :-
        read_one_term(Stream, Term, TermData, FileName),
        (
          Term == end_of_file
        ->
          PhraseList = []
        ;
          handle_term(Term,TermData),
          PhraseList = [ Term-TermData | Rest ],
          read_stream(FileName, Stream, Rest)
        )
        ;
        print_error('failed to read a term').

%% Read a file
read_file(FileName, PhraseList) :-
        already_read(FileName),
        open_stream(FileName, Stream),
        read_stream(FileName, Stream, PhraseList),
        close(Stream),!.

%% Read a list of character codes
read_codes(FileName, CharCodeList, PhraseList) :-
        already_read(FileName),
        open_chars_stream(CharCodeList, Stream),
        read_stream(FileName, Stream, PhraseList),
        close(Stream),!.

%% Reads a list of lines
read_from_lines(FileName, LineList, PhraseList) :-
        string_list_to_character_code_list(LineList, CodeList),
        read_codes(FileName,CodeList,PhraseList).

% :- dynamic loaded/1.

%% Loads a stream: same as read except that the phrases are left over
% load_stream(FileName, _) :- loaded(FileName),!.
load_stream(FileName, Stream) :-
        already_loaded(FileName),
        read_stream(FileName, Stream,_).

%% Loads a file: same as read except that the phrases are left over
% load_file(FileName) :- loaded(FileName),!.
load_file(FileName) :-
%        assertz(loaded(FileName)),
        already_loaded(FileName),
        catch(read_file(FileName, _),
              filenotfound(FileName),
              ( chop_ext(FileName,'typ',BaseName),
                  chop_ext(Source,'pl',BaseName),
                  verbose(always,[FileName,' not found, trying to check ',
                                   Source,' instead']),
                  to_read(Source)
              ) ).

%% Loads a list of char codes
% load_codes(FileName, _) :- loaded(FileName),!.
load_codes(FileName, CharCodeList) :-
        already_loaded(FileName),
        read_codes(FileName, CharCodeList, _).

%% Loads a list of lines
load_from_lines(FileName, _) :-
        chr:find_constraint(already_loaded(FileName),_),!.
load_from_lines(FileName, LineList) :-
%        assertz(loaded(FileName)),
        already_loaded(FileName),
        read_from_lines(FileName, LineList, _).

%% reinits the loaded files and the include dirs
reinit :-
        reinit_done.
%        retractall(loaded(_)),
%        retractall(include_dir(_)).
        

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% TCLP directory
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic tclp__directory/1.
tclp__directory(Dir) :-
        prolog_load_context(directory,Dir) -> true ;
        environ('SP_APP_DIR',Dir) -> true;
        Dir = unknown.

set_tclp__directory(Dir) :-
        asserta((tclp__directory(Dir) :- !)).

:- multifile user:file_search_path/2.

user:file_search_path(tclplib,Dir) :- 
        tclp__directory(D),
        concat_filename(D, 'lib', Dir).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% loading code into TCLP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tclp__consult(File) :-
        compound(File),!,
        expand_path(File,File2),
        tclp__consult(File2).
tclp__consult(File) :-
        file_exists(File),!,
        consult(File).
tclp__consult(File) :-
        filename(File,CFile),
        file_exists(CFile),!,
        (tclp__handle_goals:consult(CFile) -> %% adding tclp__handle_goals: forces reading the code in the module where it will be used
            true ;
            print_error(['Error while loading "',CFile,'"'])
        ).
tclp__consult(File) :-
        print_error(['Can''t load "',File,'"']).
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling the files to load.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handler file_loader.

constraints to_load/1, load_now/1, already_loaded/1.
constraints to_read/1, read_now/1, already_read/1.
constraints read_load_all/1.
constraints collected_data/1, collect_data/2.
constraints reinit_done/0.

delete_load @ load_now(X), already_loaded(X) <=> true.
delete_read @ read_now(X), already_read(X) <=> true.

reinit_load @ reinit_done \ already_loaded(_) <=> true.
reinit_read @ reinit_done \ already_read(_) <=> true.

load @ load_now(F) <=>
        verbose(actions,['Loading ',F,' ...']),
        load_file(F).
read @ read_now(F) <=>
        verbose(actions,['Reading ',F,' ...']),
        read_file(F, PL),
        collected_data(PL).

trigger_load @ to_load(F), read_load_all(Data) <=>
        load_now(F),
        read_load_all(Data).
trigger_read @ to_read(F), read_load_all(Data) <=>
        read_now(F),
        read_load_all(Data).

end_reading @ read_load_all(Data) <=>
        collect_data(Data,[]).

collecting_data @ collect_data(Result,Current), collected_data(PL) <=>
        append(PL,Current,New),
        collect_data(Result,New).
end_collect @ collect_data(Result,Current) <=>
        Current=Result.

add_file_to_read(FileName) :-
        to_read(FileName).
add_file_to_load(FileName) :-
        to_load(FileName).

read_and_load(PhraseList) :-
        read_load_all(PhraseList).
